/*
 * FILE --- k_fltr.cc
 *
 * PURPOSE --- Source file for a class that implements a Kalman Filter
 * algorithm for state estimation for Multiple-Input-Multiple-Output (MIMO) 
 * plants.
 * The user specifies the dimensions of the input, state, and output spaces,
 * as well as initial state estimate vector and initial state covariance 
 * matrix, and noise covariance matrices.
 * 
 * The system assumed to be in the form
 * x(k+1) = Ax(k) + Bu(k) + v_1(k)
 * y(k) = Cx(k) + v_2(k)
 * where v_1 and v_2 are zero mean (possibly gaussian) random noises
 * v_1 = system noise with autocovariance Q = E{v_1^Tv_1}
 * v_2 = output noise with autocovariance R = E{v_2^Tv_2}
 * S = E{v_1^Tv_2} is the covariance b/n system and output noises
 *
 * The Kalman filter variables are
 * xhat = the vector of estimeates of the state
 * K = Gain matrix for the filter
 * Sigma = covariance matrix for the state
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 */

// Include Files
#include "k_fltr.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif

/*************************************************************/

K_FLTR::K_FLTR(int _num_of_states, 
	       int _num_of_inputs = 1, 
	       int _num_of_outputs = 1)
{
  // No error at start
  k_fltr_err = K_FLTR_NO_ERROR;
  num_of_states = _num_of_states;
  num_of_inputs = _num_of_inputs;
  num_of_outputs = _num_of_outputs;
  
  // Create the state vector
  xhat = new matrix(num_of_states, 1);
  if (xhat->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }
  
  // Create the gain matrix
  K = new matrix(num_of_states, num_of_outputs);
  if (K->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }
  

  // Create the correlation matrix
  Sigma = new matrix(num_of_states, num_of_states);
  if (Sigma->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }

  // Create the system matrix A
  A = new matrix(num_of_states, num_of_states);
  if (A->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }
  
  // Create the system matrix B
  B = new matrix(num_of_states, num_of_inputs);
  if (B->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }

  // Create the system matrix C
  C = new matrix(num_of_outputs, num_of_states);
  if (C->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }
  
  // Create system noise autocovariance matrix
  Q = new matrix(num_of_states, num_of_states);
  if (Q->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }

  // Create output noise autocovariance matrix
  R = new matrix(num_of_outputs, num_of_outputs);
  if (R->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }
  
  // Create covariance matrix b/n system and output noise
  S = new matrix(num_of_states, num_of_outputs);
  if (S->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }

  // Create covariance vector u
  u = new matrix(num_of_inputs, 1);
  if (u->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }

  // Create covariance vector u
  y = new matrix(num_of_outputs, 1);
  if (y->matrix_err != MATRIX_NO_ERROR)
    {
      k_fltr_err = K_FLTR_MEM_ERROR;
    }

  // If any error then report
  if (k_fltr_err != K_FLTR_NO_ERROR)
    {
#ifdef NO_RCS
      printf("K_FLTR: ERROR --- Not Enought Memory!!!\n");
      printf("The Kalman Filter will not function correctly ...\n");
#else
      rcs_print("K_FLTR: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Kalman Filter will not function correctly ...\n");
#endif
    }

  // Initialize the state to zero
  xhat->set_col(0, 0.0);

  // Initialize the covariance matrix to identity
  set_cov_Sigma(1.0);

  // Initilaize the members
  output_error = 0.0;
  temp = 0.0;
}

K_FLTR::~K_FLTR()
{
  delete xhat;
  delete K;
  delete Sigma;

  delete A;
  delete B;
  delete C;

  delete Q;
  delete R;
  delete S;

  delete u;
  delete y;
}

void K_FLTR::set_state(const double *_xhat)
{
  // Set the state to the specified value
  xhat->set_col(0, _xhat);
}

void K_FLTR::get_state(double *_xhat)
{
  // Return the estimates
  for(int ii = 0; ii < num_of_states; ii++)
    {
      _xhat[ii] = xhat->get_element(ii,0);
    }
}

void K_FLTR::set_cov_Sigma(const double **_Sigma)
{
  // Set Sigma to the specified value
  Sigma->set_matrix(_Sigma);
}

void K_FLTR::set_cov_Sigma(const double *_diagonal)
{
  // Set Sigma to a diagonal matrix with specified values
  Sigma->set_matrix(0.0);
  Sigma->set_diagonal(_diagonal);
}

void K_FLTR::set_cov_Sigma(double _val)
{
  // Set Sigma to a _val*identity
  Sigma->set_matrix(0.0);
  Sigma->set_diagonal(_val);
}

void K_FLTR::get_cov_Sigma(double **_Sigma)
{
  int ii, jj; // Index variables for the loops
  
  // Return the elements of Sigma
  for(ii = 0; ii < num_of_states; ii++)
    {
      for(jj = 0; jj < num_of_states; jj++)
	{
	  _Sigma[ii][jj] = Sigma->get_element(ii,jj);
	}
    }
}

void K_FLTR::set_sys_matrix_A(const double **_A)
{
  // Set the system matrix A
  A->set_matrix(_A);
}

void K_FLTR::get_sys_matrix_A(double **_A)
{
  int ii, jj; // Index variables for the loops
  
  // Return the elements of the system matrix A
  for(ii = 0; ii < num_of_states; ii++)
    {
      for(jj = 0; jj < num_of_states; jj++)
	{
	  _A[ii][jj] = A->get_element(ii,jj);
	}
    }
}

void K_FLTR::set_sys_matrix_B(const double **_B)
{
  // Set the system matrix B
  B->set_matrix(_B);
}

void K_FLTR::set_sys_matrix_B(const double *_B)
{
  // Set the system matrix B
  B->set_matrix(_B);
}

void K_FLTR::get_sys_matrix_B(double **_B)
{
  int ii, jj; // Index variables for the loops
  
  // Return the elements of the system matrix B
  for(ii = 0; ii < num_of_states; ii++)
    {
      for(jj = 0; jj < num_of_inputs; jj++)
	{
	  _B[ii][jj] = B->get_element(ii,jj);
	}
    }
}

void K_FLTR::set_sys_matrix_C(const double **_C)
{
  // Set the system matrix C
  C->set_matrix(_C);
}

void K_FLTR::set_sys_matrix_C(const double *_C)
{
  // Set the system matrix C
  C->set_matrix(_C);
}

void K_FLTR::get_sys_matrix_C(double **_C)
{
  int ii, jj; // Index variables for the loops
  
  // Return the elements of the system matrix C
  for(ii = 0; ii < num_of_outputs; ii++)
    {
      for(jj = 0; jj < num_of_states; jj++)
	{
	  _C[ii][jj] = C->get_element(ii,jj);
	}
    }
}



void K_FLTR::set_sys_noise_autocov(const double **_Q)
{
  // Set the system noise autocovariance matrix Q
  Q->set_matrix(_Q);
}

void K_FLTR::set_sys_noise_autocov(const double *_diagonal)
{
  // Set Q to a diagonal matrix with specified values
  Q->set_matrix(0.0);
  Q->set_diagonal(_diagonal);
}

void K_FLTR::set_sys_noise_autocov(double _val)
{
  // Set Q to a _val*identity
  Q->set_matrix(0.0);
  Q->set_diagonal(_val);
}

void K_FLTR::get_sys_noise_autocov(double **_Q)
{
  int ii, jj; // Index variables for the loops
  
  // Return the system noise autocovariance matrix Q
  for(ii = 0; ii < num_of_states; ii++)
    {
      for(jj = 0; jj < num_of_states; jj++)
	{
	  _Q[ii][jj] = Q->get_element(ii,jj);
	}
    }
}

void K_FLTR::set_out_noise_autocov(const double **_R) 
{
  // Set the output noise autocovariance matrix R
  R->set_matrix(_R);
}

void K_FLTR::set_out_noise_autocov(const double *_diagonal)
{
  // Set R to a diagonal matrix with specified values
  R->set_matrix(0.0);
  R->set_diagonal(_diagonal);
}

void K_FLTR::set_out_noise_autocov(double _val)
{
  // Set R to a _val*identity
  R->set_matrix(0.0);
  R->set_diagonal(_val);
}

void K_FLTR::get_out_noise_autocov(double **_R)
{
  int ii, jj; // Index variables for the loops
  
  // Return the output noise autocovariance matrix R
  for(ii = 0; ii < num_of_outputs; ii++)
    {
      for(jj = 0; jj < num_of_outputs; jj++)
	{
	  _R[ii][jj] = R->get_element(ii,jj);
	}
    }
}


void K_FLTR::set_noise_cov(const double **_S)
{
  // Set the covariance matrix S b/n system and output noises
  S->set_matrix(_S);
}

void K_FLTR::set_noise_cov(const double *_S)
{
  // Set the covariance matrix S b/n system and output noises
  S->set_matrix(_S);
}

void K_FLTR::get_noise_cov(double **_S)
{
  int ii, jj; // Index variables for the loops
  
  // Return the covariance matrix S b/n system and output noises
  for(ii = 0; ii < num_of_states; ii++)
    {
      for(jj = 0; jj < num_of_outputs; jj++)
	{
	  _S[ii][jj] = S->get_element(ii,jj);
	}
    }
}


void K_FLTR::filter(const double *plant_outputs, 
		    const double *ctrl_inputs, 
		    double *_xhat)
{
  // Use matrix operations
  y->set_col(0, plant_outputs);
  u->set_col(0, ctrl_inputs);

  // Calculate the new gain matrix
  (*K) = ( (*A) * (*Sigma) * transpose((*C)) + (*S) ) *
    inv( (*C) * (*Sigma) * transpose((*C)) + (*R) );
  
  // Calculate the estimate of the state
  (*xhat) = (*A) * (*xhat) + (*B) * (*u) + 
    (*K) * ((*y) - (*C) * (*xhat));
  
  // Calculate the new covariance matrix
  (*Sigma) = (*A) * (*Sigma) * transpose((*A)) + (*Q) - 
    (*K) * ( (*C) * (*Sigma) * transpose((*C)) + (*R) ) * transpose((*K));
  
  // Return the new estimates
  get_state(_xhat);
}
  
